/* A simple algorithm for finding best do-loop unrolling steps. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SEARCH_RANGE	20

int dim, step;
int *guess;
float mflops;
char progname[1024], datafile[1024], prefix[1024];

#define CACHE_SIZE	4096
#define MAX_DIM		10

int computed=0, cached=0;

int cache_count=0;
struct {
  int pos[MAX_DIM];
  float mflops;
} cache[CACHE_SIZE];

static FILE *f_out=NULL;

float getmflops(int *guess);
float search(void);
void usage(void);

int main(int argc, char *argv[])
{
    int i;
    float mf0, mf;

    if (argc!=6) usage();
    dim=atoi(argv[1]);
    if (dim<=0 | dim>MAX_DIM) usage();

    strncpy(progname, argv[2], 1023);
    progname[1023]='\0';

    strncpy(datafile, argv[3], 1023);
    datafile[1023]='\0';

    if ( (step=atoi(argv[4]))<=0 ) usage();

    strncpy(prefix, argv[5], 1023);
    prefix[1023]='\0';

    {
        FILE *f_dat;
        char s[1024];

	if ((f_dat=fopen(datafile,"rt"))==NULL) {
	    fprintf(stderr,"Cannot open datafile!\n");
	    exit(1);
	}

	/* open the output file */
	sprintf(s,"%s.search.%s",progname,prefix);
	if ((f_out=fopen(s,"w+t"))==NULL) {
	    fprintf(stderr,"Cannot open output file \"%s\"!\n",s);
	    exit(1);
	}

	fprintf(f_out,"Contents of file \"%s\":\n",datafile);
	printf("Contents of file \"%s\":\n",datafile);
	do {
	    if (fgets(s, 1023, f_dat)==NULL) break;
	    fprintf(f_out,"%s",s);
	    printf("%s",s);
	} while (1);
	fclose(f_dat);
	fprintf(f_out,"Searching range: %d-%d\n",1,SEARCH_RANGE);
	printf("Searching range: %d-%d\n",1,SEARCH_RANGE);
	fprintf(f_out,"======================================================\n");
	printf("======================================================\n");
	fflush(f_out); fflush(stdout);
    }

    if ((guess=malloc(sizeof(int)*dim))==NULL) {
	fprintf(stderr,"Memory allocation error!\n");
	fclose(f_out);
	exit(1);
    }

    fprintf(f_out, "Dim=%d    Prog=%s   Data=%s   Step=%d   Prefix=%s\n",
                   dim, progname, datafile, step, prefix);
    printf("Dim=%d    Prog=%s   Data=%s   Step=%d   Prefix=%s\n",
            dim, progname, datafile, step, prefix);
    fflush(f_out); fflush(stdout);

    for (i=0; i<dim; i++) guess[i]=1;
    mf0=getmflops(guess);

    mf=search();


    fprintf(f_out,"\nNumber of evaluations: computed=%d, cache hits=%d, total=%d\n",
                   computed, cached, computed+cached);
    printf("\nNumber of evaluations: computed=%d, cache hits=%d, total=%d\n",
             computed, cached, computed+cached);

    fprintf(f_out,"\nPerformance for all unrolling steps equal to 1: %f\n",mf0);
    printf("\nPerformance for all unrolling steps equal to 1: %f\n",mf0);

    fprintf(f_out,"Best choice found: %2d",guess[0]);
    printf("Best choice found: %2d",guess[0]);

    for (i=1; i<dim; i++) {
        fprintf(f_out," %2d",guess[i]);
        printf(" %2d",guess[i]);
    }

    fprintf(f_out,"   Mflops=%f   Speedup=%f\n", mf, mf/mf0);
    printf("   Mflops=%f   Speedup=%f\n", mf, mf/mf0);

    fclose(f_out);
    return 0;
}

void usage(void)
{
    fprintf(stderr,"Usage:\n\tsearch dimensions progname datafile searching_step unroll_var_prefix\n");
    fprintf(stderr,"where dim<=%d. For example:\n", MAX_DIM);
    fprintf(stderr,"\tsearch 3 dgemm dgemm.dat 4 nn_\n");
    fprintf(stderr,"(the unrolling steps are controlled by \"nn_1\", \"nn_2\" and \"nn_3\").\n");
    exit(1);
}

int execute(char *s)
{
/*    fprintf(stderr,"%s\n",s);*/
    return system(s);
}

float getmflops(int *guess)
{
    char s[2048];
    int i,j;
    FILE *f;
    float mf;

    fprintf(f_out,"\t%2d",guess[0]);
    printf("\t%2d",guess[0]);

    for (i=1; i<dim; i++) {
	fprintf(f_out,",%2d",guess[i]);
	printf(",%2d",guess[i]);
    }

    fprintf(f_out,":   ");
    printf(":   ");

    fflush(f_out); fflush(stdout);

    for (i=0; i<dim; i++) if (guess[i]<=0 || guess[i]>SEARCH_RANGE) {
	fprintf(f_out," out of searching range.\n");
	printf(" out of searching range.\n");
	fflush(f_out); fflush(stdout);
	return -100.0;
    }

    for (j=0; j<cache_count; j++) {
	for (i=0; i<dim; i++) if (cache[j].pos[i]!=guess[i]) break;
	if (i>=dim) {
	    fprintf(f_out," mflops=%f (buffered)\n", cache[j].mflops);
	    printf(" mflops=%f (buffered)\n", cache[j].mflops);
	    fflush(f_out); fflush(stdout);
	    cached++;
	    return cache[j].mflops;
	}
    }

    computed++;
    strcpy(s,"m4 -P -D_MFLOPS_ONLY_");
    for (i=0; i<dim; i++) sprintf(s+strlen(s)," -D%s%d=%d",prefix,i+1,guess[i]);
    sprintf(s+strlen(s)," defs.m4 %s.m4 | m4post > %s.f", progname, progname);
    if (execute(s)) {
sys_err:
	fprintf(stderr,"Unable to execute: %s\n",s);
	exit(1);
    }

    sprintf(s,"rm -f %s.o; make %s >/dev/null 2>&1; rm -f %s.f", progname, progname, progname);
    if (execute(s)) {
	/* make fails, assume mf=0 */
	mf=0.0;
	goto found;
    }

    if (progname[0]=='\\') {
	sprintf(s,"%s < %s > %s.xxxx", progname, datafile, progname);
    } else {
	sprintf(s,"./%s < %s > %s.xxxx", progname, datafile, progname);
    }
    if (execute(s)) goto sys_err;

    sprintf(s,"%s.xxxx", progname);
    if ((f=fopen(s,"rt"))==NULL) {
xxxx_err:
	fprintf(stderr,"Cannot read file \"%s\"!\n",s);
	exit(1);
    }

    i=fscanf(f, "%f", &mf);
    fclose(f);
    if (i!=1) goto xxxx_err;

    sprintf(s,"rm -f %s.xxxx",progname);
    execute(s);

found:
    if (cache_count<CACHE_SIZE) {
	cache[cache_count].mflops=mf;
	for (i=0; i<dim; i++) cache[cache_count].pos[i]=guess[i];
	cache_count++;
    }

    fprintf(f_out," mflops=%f\n", mf);
    printf(" mflops=%f\n", mf);
    fflush(f_out); fflush(stdout);

    return mf;
}

float search(void)
{
    int i0, i, j, k;
    float mf0, mf;

    for (i=0; i<dim; i++) guess[i]=/*step*/1;
    mf0=getmflops(guess);
    i0=0;
    do {
	fprintf(f_out,"%2d",guess[0]);
	printf("%2d",guess[0]);
	for (i=1; i<dim; i++) {
	    fprintf(f_out," %2d", guess[i]);
	    printf(" %2d", guess[i]);
	}
	fprintf(f_out,":   %f\n", mf0);
	printf(":   %f\n", mf0);
	fflush(f_out); fflush(stdout);
	for (i=i0; ;) {
	    k=guess[i];
	    for (j=-1; j<=1; j+=2) {
		guess[i]=k+j*step;
		mf=getmflops(guess);
		if (mf0<mf || (mf0==mf && j<0)) {
		    mf0=mf; i0=i;
		    goto next;
		}
	    }
	    guess[i]=k;
	    i=(i+1)%dim;
	    if (i==i0) break;
	}
	/* reduce the step */
	if (step>1) step/=2; else break;
next:;
    } while (1);

    return mf0;
}
